package com.sd.business.service.inventory.impl;

import cn.hutool.core.util.ObjectUtil;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.ruoyi.common.constant.StatusConstant;
import com.sd.business.entity.inventory.dto.InventoryFinishedDto;
import com.sd.business.entity.inventory.dto.InventoryFinishedSelectDto;
import com.sd.business.entity.inventory.po.InventoryFinished;
import com.sd.business.entity.inventory.po.InventoryFinishedOrder;
import com.sd.business.entity.inventory.vo.InventoryFinishedVo;
import com.sd.business.entity.order.po.OrderSku;
import com.sd.business.mapper.inventory.InventoryFinishedMapper;
import com.sd.business.service.inventory.InventoryFinishedOrderService;
import com.sd.business.service.inventory.InventoryFinishedService;
import com.sd.business.service.order.OrderSkuService;
import com.sd.framework.util.Assert;
import com.sd.framework.util.sql.Sql;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.math.BigDecimal;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;


/**
 * <p>
 * 成品仓库 服务实现类
 * </p>
 *
 * @author
 * @since 2023-12-11
 */
@Service
public class InventoryFinishedServiceImpl extends ServiceImpl<InventoryFinishedMapper, InventoryFinished> implements InventoryFinishedService {

    @Autowired
    private OrderSkuService orderSkuService;

    @Autowired
    private InventoryFinishedOrderService inventoryFinishedOrderService;

    @Override
    public Page<InventoryFinishedVo> getPage(InventoryFinishedSelectDto dto) {

        Page<InventoryFinishedVo> page = Sql.create(InventoryFinishedVo.class)
                .selectAll(InventoryFinished.class)
                .from(InventoryFinished.class)
                .orderByDesc(InventoryFinished::getId)
                .page(dto);

        return page;
    }

    @Override
    public InventoryFinishedVo detail(Long id) {

        InventoryFinishedVo vo = Sql.create(InventoryFinishedVo.class)
                .selectAll(InventoryFinished.class)
                .from(InventoryFinished.class)
                .eq(InventoryFinished::getId, id)
                .one();

        Assert.notNull(vo, "未知数据");

        return vo;
    }

    @Override
    public void add(InventoryFinishedDto dto) {
        save(dto);
    }

    @Override
    public void edit(InventoryFinishedDto dto) {
        updateById(dto);
    }

    @Override
    public void delete(Long id) {
        removeById(id);
    }

    @Override
    public void productionWarehousing(List<Long> orderIdList) {
        if (orderIdList.isEmpty()) {
            return;
        }

        // 查询订单sku明细
        List<OrderSku> orderSkuList = orderSkuService.list(q -> q.in(OrderSku::getOrderId, orderIdList));
        if (orderSkuList.isEmpty()) {
            return;
        }

        // 添加成品仓库存
        List<InventoryFinished> inventoryFinishedList = orderSkuList.stream().map(item -> {
            InventoryFinished inventoryFinished = new InventoryFinished();
            inventoryFinished.setSkuSpecId(item.getSkuSpecId());
            inventoryFinished.setQuantity(item.getQuantity());
            return inventoryFinished;
        }).collect(Collectors.toList());
        inOut(inventoryFinishedList, true);

        inventoryFinishedOrderService.productionWarehousing(orderSkuList);
    }

    /**
     * 成品仓sku规格出入库
     */
    @Override
    public void inOut(List<InventoryFinished> list, boolean isIn) {

        if (list.isEmpty()) {
            return;
        }

        // 相同sku合并
        Map<Long, InventoryFinished> map = list.stream().collect(Collectors.toMap(
                InventoryFinished::getSkuSpecId,
                Function.identity(),
                (v1, v2) -> {
                    v1.setQuantity(v1.getQuantity().add(v2.getQuantity()));
                    return v1;
                }));

        synchronized (this) {

            // 获取原库存数据
            Map<Long, InventoryFinished> oldStorageMap = mapKEntity(InventoryFinished::getSkuSpecId,
                    q -> q.in(InventoryFinished::getSkuSpecId, map.keySet()));

            // 计算入库后库存数量
            for (InventoryFinished item : map.values()) {
                InventoryFinished inventoryFinished = oldStorageMap.get(item.getSkuSpecId());
                if (inventoryFinished == null) {
                    continue;
                }
                BigDecimal quantity;
                if (isIn) {
                    quantity = inventoryFinished.getQuantity().add(item.getQuantity());
                } else {
                    quantity = inventoryFinished.getQuantity().subtract(item.getQuantity());
                }
                item.setId(inventoryFinished.getId());
                item.setQuantity(quantity);
            }

            // 更新数据库
            saveOrUpdateBatch(map.values());
        }

    }

    @Override
    public void removeOrder(Long orderId) {

        // 查询成品仓库
        List<InventoryFinishedOrder> list = inventoryFinishedOrderService.list(q -> q
                .eq(InventoryFinishedOrder::getOrderInfoId, orderId)
                .eq(InventoryFinishedOrder::getStatus, StatusConstant.YES));

        if (list.isEmpty()) {
            return;
        }

        // 成品库存出库
        outByFinishedOrder(list);

        // 删除成品仓库订单以及明细
        inventoryFinishedOrderService.removeByOrderId(orderId);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public boolean lockNoSourceStorage(List<? extends OrderSku> orderSkuList) {
        if (ObjectUtil.isEmpty(orderSkuList)) {
            return true;
        }

        Map<Long, BigDecimal> map = orderSkuList.stream()
                .collect(Collectors.toMap(OrderSku::getSkuSpecId, OrderSku::getQuantity, BigDecimal::add));

        synchronized (this) {

            List<InventoryFinished> list = list(q -> q.in(InventoryFinished::getSkuSpecId, map.keySet()));

            if (map.size() != list.size()) {
                return false;
            }

            // 查询无源成品数据
            List<InventoryFinishedOrder> inventoryFinishedOrderList = inventoryFinishedOrderService.list(q -> q
                    .in(InventoryFinishedOrder::getSkuSpecId, map.keySet())
                    .eq(InventoryFinishedOrder::getStatus, 1)
                    .isNull(InventoryFinishedOrder::getOrderInfoId));
            Map<Long, BigDecimal> inventoryFinishedMap = inventoryFinishedOrderList.stream()
                    .collect(Collectors.toMap(
                            InventoryFinishedOrder::getSkuSpecId,
                            InventoryFinishedOrder::getExistingQuantity,
                            BigDecimal::add));

            if (map.size() != inventoryFinishedMap.size()) {
                return false;
            }

            for (InventoryFinished inventoryFinished : list) {
                BigDecimal lockQuantity = map.get(inventoryFinished.getSkuSpecId());
                // 查询成品sku现存数量，减去锁定库存，得出当前可用数量
                BigDecimal existingQuantity = inventoryFinishedMap.getOrDefault(inventoryFinished.getSkuSpecId(), BigDecimal.ZERO);
                if (existingQuantity.subtract(inventoryFinished.getLockQuantity()).compareTo(lockQuantity) >= 0) {
                    inventoryFinished.setLockQuantity(inventoryFinished.getLockQuantity().add(lockQuantity));
                } else {
                    return false;
                }
            }

            updateBatchById(list);
        }

        return true;
    }

    /**
     * 成品库存出库
     *
     * @param list 成品仓库
     */
    private void outByFinishedOrder(List<InventoryFinishedOrder> list) {

        List<InventoryFinished> inventoryFinishedList = list.stream().map(item -> {
            InventoryFinished inventoryFinished = new InventoryFinished();
            inventoryFinished.setQuantity(item.getQuantity());
            inventoryFinished.setSkuSpecId(item.getSkuSpecId());
            return inventoryFinished;
        }).collect(Collectors.toList());

        inOut(inventoryFinishedList, false);

    }

}
